@@ -83,7 +83,8 @@ def train_from_folder(
     evaluate_every = 1000,
     generate = False,
     generate_interpolation = False,
-    hamburger_res_layers = [],
+    hamburger_res_layers = [32],
+    use_evonorm = False,
     interpolation_num_steps = 100,
     save_frames = False,
     num_image_tiles = 8,
@@ -103,6 +104,7 @@ def train_from_folder(
         batch_size = batch_size,
         gradient_accumulate_every = gradient_accumulate_every,
         hamburger_res_layers = cast_list(hamburger_res_layers),
+        use_evonorm = use_evonorm,
         image_size = image_size,
         fmap_max = fmap_max,
         transparent = transparent,
